#include <stdio.h>
#include <stdint.h>
#include <pthread.h>
#include <stdlib.h>
#include <math.h>

//#define DO_NOTHING

#define SAMPLES 64

static int verbose;

#define R do { printf("%s:%d: err\n", __FILE__, __LINE__); exit(1); } while (0)
#define V(...) do { if (verbose) printf(__VA_ARGS__); } while (0)

#include "alsa.c"
#include "gui.c"
#include "input.c"

void *jack_client_open(const char *cname, int dummy, int *status)
{
  static int id = 0;
  void *ret;
  if (getenv("MINIJACK_VERBOSE")) verbose = 1;
  V("jack_client_open %s\n", cname);
  *status = 0;
  id++;
  ret = (void *)(intptr_t)id;
  return ret;
}

int jack_is_realtime(void *dummy)
{
  V("jack_is_realtime %d\n", (int)(intptr_t)dummy);
  return 0;
}

uint32_t jack_get_sample_rate(void *dummy)
{
  V("jack_get_sample_rate %d\n", (int)(intptr_t)dummy);
  return 44100;
}

uint32_t jack_get_buffer_size(void *dummy)
{
  V("jack_get_buffer_size %d\n", (int)(intptr_t)dummy);
  return SAMPLES;
}

int jack_set_xrun_callback(void *a, void *b, void *c)
{
  V("jack_set_xrun_callback %d\n", (int)(intptr_t)a);
  return 0;
}

int jack_set_sample_rate_callback(void *a, void *b, void *c)
{
  V("jack_set_sample_rate_callback %d\n", (int)(intptr_t)a);
  return 0;
}

void jack_on_shutdown(void *dummy, void *a, void *b)
{
  V("jack_on_shutdown %d\n", (int)(intptr_t)dummy);
}

int jack_set_buffer_size_callback(void *dummy, void *a, void *b)
{
  V("jack_set_buffer_size_callback %d\n", (int)(intptr_t)dummy);
  return 0;
}

static int (*cb_amp)(uint32_t, void *);
static void *cbarg_amp;
static int (*cb_fx)(uint32_t, void *);
static void *cbarg_fx;

int jack_set_process_callback(void *dummy, void *a, void *b)
{
  V("jack_set_process_callback %d\n", (int)(intptr_t)dummy);
  if ((int)(intptr_t)dummy == 1) { cb_amp = a; cbarg_amp = b; }
  else if ((int)(intptr_t)dummy == 2) { cb_fx = a; cbarg_fx = b; }
  else { V("unknown client %d\n", (int)(intptr_t)dummy); exit(1); }
  return 0;
}

int jack_set_port_registration_callback(void *dummy, void *a, void *b)
{
  V("jack_set_port_registration_callback %d\n", (int)(intptr_t)dummy);
  return 0;
}

static void (*con_cb)(uint32_t, uint32_t, int, void *);

int jack_set_port_connect_callback(void *dummy, void *a, void *b)
{
  V("jack_set_port_connect_callback %d\n", (int)(intptr_t)dummy);
  con_cb = a;
  return 0;
}

void *jack_port_register(void *id, char *n1, char *n2, unsigned long f,
                         unsigned long bufsize)
{
  static int ret = 0;
  V("jack_port_register (id %d) '%s' '%s' %ld (return %d)\n",
         (int)(intptr_t)id, n1, n2,
         bufsize, ret+1);
  ret++;
  return (void *)(intptr_t)ret;
}

static float inbuf[SAMPLES];
static float midbuf[SAMPLES];
static float outbufleft[SAMPLES];
static float outbufright[SAMPLES];
static float dumbuf[SAMPLES];

static unsigned char out[SAMPLES*4];

static void *go(void *u)
{
#ifndef DO_NOTHING
  int turn = 0;
  FILE *f;
#endif
  unsigned char *bu;
  int i;
  char filename[256];

  init_gui();

start:
  init_input();
  memset(inbuf, 0, SAMPLES*sizeof(float));
  while (gui_state == GUI_START) {
    cb_amp(SAMPLES, cbarg_amp);
    cb_fx(SAMPLES, cbarg_fx);
    usleep(100000);
  }

  /* let's go */
#ifndef DO_NOTHING
  while (1) {
    turn++;
    sprintf(filename, "gx-head-replay-%5.5d.raw", turn);
    f = fopen(filename, "r"); if (!f) break;
    fclose(f);
  }
  f = fopen(filename, "w"); if (!f) R;
#endif
  open_alsa();
  play_sound(out);

  while (gui_state == GUI_GO) {
    read_input(inbuf);
    cb_amp(SAMPLES, cbarg_amp);
    cb_fx(SAMPLES, cbarg_fx);
#ifndef DO_NOTHING
    if (fwrite(outbufleft, SAMPLES * sizeof(float), 1, f) != 1) R;
#endif
    bu = out;
    for (i = 0; i < SAMPLES; i++, bu+=4) {
      int v;
      v = outbufleft[i] * 32767;
      if (v > 32767) v = 32767;
      if (v < -32768) v = -32768;
      bu[0] = v & 255;
      bu[1] = v >> 8;
      bu[2] = v & 255;
      bu[3] = v >> 8;
    }
//    { int n = out64(); if (n>100) { fprintf(stderr, "-"); fflush(stderr); play_sound(silence); } }
    play_sound(out);
  }

#ifndef DO_NOTHING
  if (fclose(f)) R;
#endif
  close_alsa();
  close_input();

  goto start;

  return 0;
}

static pthread_t thid;

int jack_activate(void *dummy)
{
  V("jack_activate %d\n", (int)(intptr_t)dummy);
  if ((int)(intptr_t)dummy == 2)
    if (pthread_create(&thid, NULL, go, NULL)) abort();
  return 0;
}

char *jack_port_name(void *p)
{
  static char t[128];
  sprintf(t, "port %p", p);
  V("jack_port_name %d\n", (int)(intptr_t)p);
//  return "a";
  return t;
}

int jack_connect(void *dummy, char *a, char *b)
{
  V("jack_connect %d %p (%s) %p (%s)\n", (int)(intptr_t)dummy, a, a, b, b);
  con_cb((uint32_t)(uintptr_t)a, (uint32_t)(uintptr_t)b, 1, 0);
  return 0;
}

void *jack_port_by_id(void *client, uint32_t port)
{
  V("jack_port_by_id %d\n", port);
  return (void *)(uintptr_t)port;
}

void **jack_port_get_connections(void *p)
{
  void **t = malloc(sizeof(void *));
  V("jack_port_get_connections %p\n", p);
  *t = 0;
  return t;
}

int jack_port_unregister(void *a, void *b)
{
  V("jack_port_unregister %p %p\n", a, b);
  return 0;
}

int jack_deactivate(void *client)
{
  V("jack_deactivate %p\n", client);
  return 0;
}

int jack_client_close(void *client)
{
  V("jack_client_close %p\n", client);
  return 0;
}

pthread_t jack_client_thread_id(void *dummy)
{
  V("jack_client_thread_id %d\n", (int)(intptr_t)dummy);
  return thid;
}

void *jack_port_get_buffer(void *u, uint32_t n)
{
  int p = (int)(intptr_t)u;
  V("jack_port_get_buffer %d %u\n", p, n);
  if (p == 2) {
#if 0
    static unsigned long p=0;
    unsigned long i;
    //for (i = 0; i < 1024; i++, p++) inbuf[i] = (float)rand()/(float)RAND_MAX; //cos(p*440.*2*M_PI/44100);
    for (i = 0; i < 1024; i++, p++) inbuf[i] = cos(p*110.*2*M_PI/44100);
#endif
    return inbuf;
  }
  if (p == 5) return midbuf;
  if (p == 3) return midbuf;
  if (p == 6) return outbufleft;
  if (p == 7) return outbufright;
  if (p == 1 || p == 4) return dumbuf;
  V("error\n");
  exit(1);
}

int jack_midi_get_event_count(void *u)
{
  V("jack_midi_get_event_count %d\n", (int)(intptr_t)u);
  return 0;
}

void jack_midi_clear_buffer(void *u)
{
  V("jack_midi_clear_buffer %d\n", (int)(intptr_t)u);
}

float jack_cpu_load(void *client)
{
  return 1.;
}

uint32_t jack_frame_time(void *u)
{
  return 0;
}
